SuperStrict

Import chaos.mathext

Const FRACTAL_SQUARED:Int	= 1 'Fractal types - CUBIC is currently experimental
Const FRACTAL_CUBIC:Int	= 2

Type TCamera Extends TEntity
	Field AOV:Float				'Angle of view
	Field RenderShadows:Int		'Render shadows Yes/No
	Field AntiAliasing:Int			'4x Antialiasing Yes/No
	Field BackgroundColor:TVector	'Backgroundcolor
	
	Field JuliaSet:TJuliaSet		'Fractal, that is rendered
	Field LightPosition:TVector		'Light position
	
	Method SetJuliaSet( JuliaSet:TJuliaSet )
		Self.JuliaSet = JuliaSet
	End Method
	
	Method SetBackgroundColor( Color:TVector )
		Self.BackgroundColor = Color
	End Method
	
	Method SetLightPosition( LightPosition:TVector )
		Self.LightPosition = LightPosition
	End Method
	
	Method SetShadowFlag( Flag:Int )
		Self.RenderShadows = Flag <> 0
	End Method
	
	Method SetAntiAliasing( Flag:Int )
		Self.AntiAliasing = Flag <> 0
	End Method
	
	Method Render( ScreenWidth:Int, ScreenHeight:Int, Destination:TPixmap, FeedbackFunc( Percentage:Int ), FeedbackDelay:Int )
		Local PlaneDist:Float = 1/Tan( Self.AOV/2.0 ) 'Camery distance from the projection plane
		
		Local AspectRatio:Float = ScreenHeight/Float( ScreenWidth ) 'The aspect ratio
		
		Local RayPosition:TVector = TFormPoint( Vec3( 0, 0, 0 ), Self, Self.JuliaSet ) 'Camera position in the fractal space
		Local LightPosition:TVector = TFormPoint( Self.LightPosition, Self, Self.JuliaSet ) 'Same for the light position
		
		Local FeedbackTimer:Int = MilliSecs()
		
		For Local X:Float = 0 To ScreenWidth - 1
			For Local Y:Float = 0 To ScreenHeight - 1
				Local PixelColor:TVector
				
				If Self.AntiAliasing Then 'If antialiasing is set to true, sample four times, otherwise only once
					Local Color1:TVector = SamplePixel( X - 0.25, Y - 0.25, RayPosition, LightPosition, ScreenWidth, ScreenHeight, AspectRatio, PlaneDist )
					Local Color2:TVector = SamplePixel( X + 0.25, Y - 0.25, RayPosition, LightPosition, ScreenWidth, ScreenHeight, AspectRatio, PlaneDist )
					Local Color3:TVector = SamplePixel( X + 0.25, Y + 0.25, RayPosition, LightPosition, ScreenWidth, ScreenHeight, AspectRatio, PlaneDist )
					Local Color4:TVector = SamplePixel( X - 0.25, Y + 0.25, RayPosition, LightPosition, ScreenWidth, ScreenHeight, AspectRatio, PlaneDist )
					
					PixelColor = Color1.Add( Color2 ).Add( Color3 ).Add( Color4 ).Divide( 4.0 ) 'Calculate average color
				Else
					PixelColor = SamplePixel( X, Y, RayPosition, LightPosition, ScreenWidth, ScreenHeight, AspectRatio, PlaneDist )
				EndIf
				
				WritePixel( Destination, X, Y, PixelColor.ToRGB() | $FF000000 )
				
				If MilliSecs() - FeedbackTimer > FeedbackDelay Then 'Call the FeedbackFunction in certain intervals
					Local Percentage:Int = Floor( ( X + Y/Float( ScreenHeight ) )/Float( ScreenWidth )*100 )
					
					If Not FeedbackFunc( Percentage ) Then 'If the FeedbackFunction returns false, exit
						Destination.ClearPixels( BackgroundColor.ToRGB() )
						
						Return
					EndIf
					
					FeedbackTimer = MilliSecs()
				EndIf
			Next
		Next
	End Method
	
	Method SamplePixel:TVector( X:Float, Y:Float, RayPosition:TVector, LightPosition:TVector, ScreenWidth:Int, ScreenHeight:Int, AspectRatio:Float, PlaneDist:Float )
		'Calculate direction of the light ray, transform it in the fractal space and normalize it
		Local Ray:TVector = Vec3( X/ScreenWidth - 0.5, ( Y/ScreenHeight - 0.5 )*AspectRatio, PlaneDist )
		
		Ray = TFormNormal( Ray, Self, Self.JuliaSet )
		
		'Set the ray position on the bounding sphere to speed up the rendering process
		Local Intersection:TVector = Self.JuliaSet.BoundingSphereIntersection( RayPosition, Ray )
		
		Local Output:TVector = Self.BackgroundColor.Copy() 'Set ray color to background color
		
		If Intersection Then
			Local Distance:Float = Self.JuliaSet.IntersectRay( Intersection, Ray ) 'Intersect ray with fractal
			
			'Intersection? If yes, calculate color
			If Distance < Self.JuliaSet.Epsilon Then Output = Self.CalculateColor( LightPosition, RayPosition, Intersection )
		EndIf
		
		Return Output
	End Method
	
	Method CalculateColor:TVector( LightPosition:TVector, RayPosition:TVector, Intersection:TVector )
		 'Calculate color according to the phong lighting model
		Local Diffuse:TVector	= Self.JuliaSet.DiffuseColor.Copy()
		Local SpecularH:Float	= Self.JuliaSet.SpecularHighlight
		Local Shading:Float	= Self.JuliaSet.ShadingExponent
		
		Local Normal:TVector			= Self.JuliaSet.GetNormal( Intersection )
		Local LightVector:TVector		= LightPosition.Copy().Substract( Intersection ).Normalize()
		Local EyeVector:TVector		= RayPosition.Copy().Substract( Intersection ).Normalize()
		Local NormalLightDotP:Float		= LightVector.DotP( Normal )
		Local ReflectedVector:TVector	= LightVector.Copy().Substract( Normal.Multiply( 2*NormalLightDotP ) )
		
		'Diffuse.Add( Normal.Absolute().Multiply( 0.3*255 ) )
		
		Local SpecularScalar:Float = ( SpecularH*Max( EyeVector.DotP( ReflectedVector.Multiply( -1 ) ), 0 )^Shading )*255
		
		Diffuse.Multiply( Max( NormalLightDotP, 0 ) ).Add( Vec3( SpecularScalar, SpecularScalar, SpecularScalar ) )
		
		If Self.RenderShadows Then 'If shadows are enabled, send a second ray towards the light source
			Local RayStart:TVector = Intersection.Copy().Add( LightVector.Copy().Multiply( Self.JuliaSet.Epsilon*10 ) )
			
			Local Distance:Float = Self.JuliaSet.IntersectRay( RayStart, LightVector )
			
			'If the ray intersects with the fractal, then this point cannot be seen from the light - ergo is shadowed
			If Distance < Self.JuliaSet.Epsilon Then Diffuse.Multiply( 0.3 ) 'Darken the color a bit, to simulate shadows
		EndIf
		
		If Diffuse.X > 255 Then Diffuse.X = 255 'Clamp color to valid values
		If Diffuse.Y > 255 Then Diffuse.Y = 255
		If Diffuse.Z > 255 Then Diffuse.Z = 255
		If Diffuse.X < 0 Then Diffuse.X = 0
		If Diffuse.Y < 0 Then Diffuse.Y = 0
		If Diffuse.Z < 0 Then Diffuse.Z = 0
		
		Return Diffuse
	End Method
	
	Function Create:TCamera( AOV:Float = 60 )
		Local Cam:TCamera = New TCamera
			Cam.AOV = AOV
			Cam.SetBackgroundColor( Vec4( 255, 255, 255, 1 ) )
			Cam.SetLightPosition( Vec3( 3, 5, -3 ) )
		
		Return Cam
	End Function
End Type

Type TJuliaSet Extends TEntity
	Const BOUNDING_SPHERE:Float = 3.0 'Radius of the bounding sphere
	Const DIVERGENCE_TRESHOLD:Float = 10.0 'Treshold to determine, whether a point is diverging or converging
	Const NORMAL_DELTA:Float = 0.0001 'Accuracy of the normal estimator
	
	Field Iterations:Int 'Number of iterations used to estimate the convergence of a point
	Field Epsilon:Float 'Treshold of the distance estimator
	Field FractalType:Int 'Squared/cubic
	
	Field DiffuseColor:TVector 'Lighting data
	Field ShadingExponent:Float
	Field SpecularHighlight:Float
	
	Field C:TVector 'Fractal constant
	
	Method New()
		FractalType = FRACTAL_SQUARED
	End Method
	
	Method SetC( C:TVector )
		Self.C = C
	End Method
	
	Method SetApproximation( Iterations:Int, Epsilon:Float )
		Self.Iterations = Iterations
		Self.Epsilon = Epsilon
	End Method
	
	Method SetMaterial( Diffuse:TVector, Shading:Float = -1, Specular:Float = -1 )
		Self.DiffuseColor = Diffuse
		If Shading  >= 0 Then Self.ShadingExponent = Shading
		If Specular >= 0 Then Self.SpecularHighlight = Specular
	End Method
	
	Method IntersectRay:Float( RayPosition:TVector, Ray:TVector )
		While True
			Local Z:TVector = RayPosition.Copy() 'The quaternion in the quaternion 4D - space
			Local Zd:TVector = Vec4( 1, 0, 0, 0 ) 'Derivation of Z
			
			Z.W = 0 'W would be 1 for a normal 3D - Vector, but should be initially zero for this quaternion
			
			For Local I:Int = 1 To Self.Iterations 'Estimate divergence/convergence
				If FractalType = FRACTAL_SQUARED Then
					Zd = Z.Copy().QuaternionMultiply( Zd ).Multiply4( 2.0 )
					Z = Z.QuaternionSquare().Add4( Self.C )
				Else
					Zd = Z.Copy().QuaternionSquare().QuaternionMultiply( Zd ).Multiply4( 2.0 )
					Z = Z.Copy().QuaternionSquare().QuaternionMultiply( Z ).Add4( Self.C )
				EndIf
				
				If Z.DotP4( Z ) > DIVERGENCE_TRESHOLD Then Exit 'Equal to Length4 without the square root
			Next
			
			Local LengthZ:Float = Z.Length4()
			
			'Distance estimator for the uncounding volume. Formula and further explanations, see: http://graphics.cs.uiuc.edu/~jch/papers/rtqjs.pdf
			Local UnboundDistance:Float = 0.5*LengthZ*Log10( LengthZ )/Zd.Length4()
			
			RayPosition.Add( Ray.Copy().Multiply( UnboundDistance ) )
			
			If UnboundDistance < Self.Epsilon Or RayPosition.DotP( RayPosition ) > BOUNDING_SPHERE Then Return UnboundDistance
		Wend
	End Method
	
	Method BoundingSphereIntersection:TVector( RayPosition:TVector, Ray:TVector ) 'Fast Ray-Sphere-Intersection. Source: http://www.devmaster.net/wiki/Ray-sphere_intersection
		Local B:Float = 2*RayPosition.DotP( Ray )
		Local C:Float = RayPosition.DotP( RayPosition ) - Self.BOUNDING_SPHERE
		Local D:Float = Sqr( B*B - 4*C )
		
		If IsNan( D ) Then Return Null
		
		Local E0:Float = ( -B + D )*0.5
		Local E1:Float = ( -B - D )*0.5
		
		Return RayPosition.Copy().Add( Ray.Copy().Multiply( Min( E0, E1 ) ) )
	End Method
	
	Method GetNormal:TVector( Position:TVector ) 'Normalenestimator
		Position = Position.Copy(); Position.W = 0
		
		'Basic principle: Six points near the point to be examined are chosen and then Julia-Transformed.
		'Based on the position difference the normal is then computed
		'Source: http://www.devmaster.net/forums/showthread.php?t=4448
		
		Local GradientX0:TVector = Position.Copy().Substract4( Vec4( NORMAL_DELTA, 0, 0, 0 ) )
		Local GradientX1:TVector = Position.Copy().Add4(       Vec4( NORMAL_DELTA, 0, 0, 0 ) )
		Local GradientY0:TVector = Position.Copy().Substract4( Vec4( 0, NORMAL_DELTA, 0, 0 ) )
		Local GradientY1:TVector = Position.Copy().Add4(       Vec4( 0, NORMAL_DELTA, 0, 0 ) )
		Local GradientZ0:TVector = Position.Copy().Substract4( Vec4( 0, 0, NORMAL_DELTA, 0 ) )
		Local GradientZ1:TVector = Position.Copy().Add4(       Vec4( 0, 0, NORMAL_DELTA, 0 ) )
		
		If FractalType = FRACTAL_SQUARED Then
			For Local I:Int = 1 To Self.Iterations
				GradientX0 = GradientX0.QuaternionSquare().Add4( Self.C )
				GradientX1 = GradientX1.QuaternionSquare().Add4( Self.C )
				GradientY0 = GradientY0.QuaternionSquare().Add4( Self.C )
				GradientY1 = GradientY1.QuaternionSquare().Add4( Self.C )
				GradientZ0 = GradientZ0.QuaternionSquare().Add4( Self.C )
				GradientZ1 = GradientZ1.QuaternionSquare().Add4( Self.C )
			Next
		Else
			For Local I:Int = 1 To Self.Iterations
				GradientX0 = GradientX0.Copy().QuaternionSquare().QuaternionMultiply( GradientX0 ).Add4( Self.C )
				GradientX1 = GradientX1.Copy().QuaternionSquare().QuaternionMultiply( GradientX0 ).Add4( Self.C )
				GradientY0 = GradientY0.Copy().QuaternionSquare().QuaternionMultiply( GradientX0 ).Add4( Self.C )
				GradientY1 = GradientY1.Copy().QuaternionSquare().QuaternionMultiply( GradientX0 ).Add4( Self.C )
				GradientZ0 = GradientZ0.Copy().QuaternionSquare().QuaternionMultiply( GradientX0 ).Add4( Self.C )
				GradientZ1 = GradientZ1.Copy().QuaternionSquare().QuaternionMultiply( GradientX0 ).Add4( Self.C )
			Next
		EndIf
		
		Local GradientX:Float = GradientX1.Length4() - GradientX0.Length4()
		Local GradientY:Float = GradientY1.Length4() - GradientY0.Length4()
		Local GradientZ:Float = GradientZ1.Length4() - GradientZ0.Length4()
		
		Return Vec3( GradientX, GradientY, GradientZ ).Normalize()
	End Method
	
	Function Create:TJuliaSet()
		Return New TJuliaSet
	End Function
End Type
